/* i4m.c */
/* matrix operations using modular arithmetic */

#include <stdio.h>
#include "integer.h"
#include "fun.h"

MPMATm *SWAP_ROWSm(USI p, USI q, MPMATm *Mptr)
/*
 * interchange rows p and q of *Mptr.
 */
{
	unsigned int j, x;
	MPMATm *Nptr;

	Nptr = COPYMATm(Mptr);
	for (j = 0; j <= Nptr->C - 1; j++)
	{
		x = Nptr->V[p][j];
		Nptr->V[p][j] = Nptr->V[q][j];
		Nptr->V[q][j] = x;
	}
	return (Nptr);
}

MPMATm *SWAP_COLSm(USI p, USI q, MPMATm *Mptr)
/*
 * interchange columns p and q of *Mptr.
 */
{
	unsigned int i, x;
	MPMATm *Nptr;

	Nptr = COPYMATm(Mptr);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		x = Nptr->V[i][p];
		Nptr->V[i][p] = Nptr->V[i][q];
		Nptr->V[i][q] = x;
	}
	return (Nptr);
}

MPMATm *SCALAR_ROWm(USI p, USI a, MPMATm *Mptr, USI m)
/*
 * multiplying the pth row of *Mptr by a (mod m).
 */
{
	unsigned int j;
	MPMATm *Nptr;

	Nptr = COPYMATm(Mptr);
	for (j = 0; j <= Nptr->C - 1; j++)
		Nptr->V[p][j] = MULTm(Nptr->V[p][j], a, m);
	return(Nptr);
}


MPMATm *SCALAR_COLm(USI p, USI a, MPMATm *Mptr, USI m)
/*
 * multiplying the pth column of *Mptr by a (mod m).
 */
{
	unsigned int i;
	MPMATm *Nptr;

	Nptr = COPYMATm(Mptr);
	for (i = 0; i <= Nptr->R - 1; i++)
		Nptr->V[i][p] = MULTm(Nptr->V[i][p], a, m);
	return(Nptr);
}

MPMATm *ADD_MULT_ROWm(USI p, USI q, USI a, MPMATm *Mptr, USI m)
/*
 * adding a times row p to row q of *Mptr (mod m).
 */
{
	unsigned int j, x;
	MPMATm *Nptr;

	Nptr = COPYMATm(Mptr);
	for (j = 0; j <= Nptr->C - 1; j++)
	{
		x = MULTm(Nptr->V[p][j], a, m);
		Nptr->V[q][j] = ADDm(Nptr->V[q][j], x, m);
	}
	return (Nptr);
}

MPMATm *ADD_MULT_COLm(USI p, USI q, USI a, MPMATm *Mptr, USI m)
/*
 * adding a times column p to column q of *Mptr (mod m).
 */
{
	unsigned int i, x;
	MPMATm *Nptr;

	Nptr = COPYMATm(Mptr);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		x = MULTm(Nptr->V[i][p], a, m);
		Nptr->V[i][q] = ADDm(Nptr->V[i][q], x, m);
	}
	return (Nptr);
}

MPMATm *TRANSPOSEm(MPMATm *Mptr)
/*
 * returns the transpose of *Mptr.
 */
{
	MPMATm *Nptr;
	unsigned int i, j;

	Nptr = BUILDMATm(Mptr->C, Mptr->R);
	for (j = 0; j <= Nptr->R - 1; j++)
	{
		for (i = 0; i <= Nptr->C - 1; i++)
			Nptr->V[j][i] = Mptr->V[i][j];
	}
	return (Nptr);
}

MPMATm *DELETE_ROWm(USI r, MPMATm *Mptr)
/*
 * deletes row r of *Mptr.
 */
{
	unsigned int i, j;
	MPMATm *Nptr;

	r--;
	Nptr = BUILDMATm(Mptr->R - 1, Mptr->C);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		if (i < r)
		{
			for (j = 0; j <= Nptr->C - 1; j++)
				Nptr->V[i][j] = Mptr->V[i][j];
		}
		else
		{
			for (j = 0; j <= Nptr->C - 1; j++)
				Nptr->V[i][j] = Mptr->V[i + 1][j];
		}
	}
	return (Nptr);
}

MPMATm *DELETE_COLm(USI c, MPMATm *Mptr)
/*
 * deletes  column c of *Mptr.
 */
{
	unsigned int i, j;
	MPMATm *Nptr;

	c--;
	Nptr = BUILDMATm(Mptr->R, Mptr->C - 1);
	for (j = 0; j <= Nptr->C - 1; j++)
	{
		if (j < c)
		{
			for (i = 0; i <= Nptr->R - 1; i++)
				Nptr->V[i][j] = Mptr->V[i][j];
		}
		else
		{
			for (i = 0; i <= Nptr->R - 1; i++)
				Nptr->V[i][j] = Mptr->V[i][j + 1];
		}
	}
	return (Nptr);
}

MPMATm *KRONECKERm(MPMATm *Mptr, MPMATm *Nptr, USI m)
/*
 * returns the Kronecker product of *Mptr and *Nptr (mod m).
 */
{
	unsigned int i1, i2, j1, j2, m1, m2, n1, n2;
	MPMATm *Qptr;

	m1 = Mptr->R;
	m2 = Nptr->R;
	n1 = Mptr->C;
	n2 = Nptr->C;
	Qptr = BUILDMATm(m1 * m2, n1 * n2);
	for (i1 = 0; i1 <= m1 - 1; i1++)
	{
		for (j1 = 0; j1 <= n1 - 1; j1++)
		{
			for (i2 = 0; i2 <= m2 - 1; i2++)
			{
				for (j2 = 0; j2 <= n2 - 1; j2++)
					Qptr->V[m2 * i1 + i2][n2 * j1 + j2] = MULTm(Mptr->V[i1][j1], Nptr->V[i2][j2], m);
			}
		}
	}
	return(Qptr);
}

MPMATm *DIRECT_SUMm(MPMATm *Mptr, MPMATm *Nptr)
/*
 * returns the direct sum of *Mptr and *Nptr.
 */
{
	unsigned int i, j, m1, m2, n1, n2;
	MPMATm *Qptr;
	
	m1 = Mptr->R;
	m2 = Nptr->R;
	n1 = Mptr->C;
	n2 = Nptr->C;
	Qptr = BUILDMATm(m1 + m2, n1 + n2);
	for (i = 0; i <= m1 + m2 - 1; i++)
	{
		for (j = 0; j <= n1 + n2 - 1; j++)
		{
			if (i <= m1 - 1 && j <= n1 - 1)
				Qptr->V[i][j] = Mptr->V[i][j];
			else if (m1 <= i && n1 <= j)
				Qptr->V[i][j] = Nptr->V[i - m1][j - n1];
			else
				Qptr->V[i][j] = 0;
		}
	}
	return (Qptr);
}

MPMATm *JORDANm(USI n, USI a)
/*
 * returns the elementary Jordan matrix of size n, diagonal elt. = a.
 */
{
	unsigned int i, j;
	MPMATm *Mptr;
	
	Mptr = BUILDMATm(n, n);
	for (i = 0; i <= n - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
		{
			if (i == j)
				Mptr->V[i][j] = a;
			else if (i == j + 1)
				Mptr->V[i][j] = 1;
			else
				Mptr->V[i][j] = 0;
		}
	}
	return (Mptr);
}

MPMATm *BANDMATm(USI a, USI b, USI k, USI n)
/*
 * Constructs the n x n band matrix with a on the diagonal and b
 * on the k sub and super diagonals. Here n >= 2 and 1 <= k <= n - 1.
 */
{
	unsigned int i, j;
	MPMATm *Mptr;

	Mptr = ZEROMNm(n, n);
	for (i = 0; i < n; i++)
		Mptr->V[i][i] = a;
	for (j = 1; j <= k; j++)
	{
		for (i = 0; i <= n - 1 - j; i++)
		{
			Mptr->V[i][i + j] = b;
			Mptr->V[i + j][i] = b;
		}
	}
	return Mptr;
}

MPMATm *IDENTITYm(USI n)
/*
 * returns the identity matrix of size n.
 */
{
	unsigned int i, j;
	MPMATm *Mptr;

	Mptr = BUILDMATm(n, n);
	for (i = 0; i <= n - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
		{
			if ( i == j)
				Mptr->V[i][j] = 1;
			else
				Mptr->V[i][j] = 0;
		}
	}
	return (Mptr);
}

MPMATm *ZEROMNm(USI m, USI n)
/*
 * returns the zero  m x n matrix.
 */
{
	unsigned int i, j;
	MPMATm *Mptr;

	Mptr = BUILDMATm(m, n);
	for (i = 0; i <= m - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
			Mptr->V[i][j] = 0;
	}
	return (Mptr);
}

MPMATm *SUBMATROWSm(MPMATm *Mptr, USI *Cptr, USI n)
/*
 * Cptr is an array of n (distinct) non-negative integers, n < Mptr->R.
 * Output: the submatrix formed from rows Cptr[0],...,Cptr[n-1].
 */
{
	unsigned int i, j;
	MPMATm *Nptr;

	Nptr = BUILDMATm(n, Mptr->C);
	for (i = 0; i < n; i++)
		for (j = 0; j < Mptr->C; j++)
			Nptr->V[i][j] = Mptr->V[Cptr[i]][j];
	return Nptr;
}

MPMATm *SUBMATCOLUMNSm(MPMATm *Mptr, USI *Cptr, USI n)
/*
 * Cptr is an array of n (distinct) non-negative integers, n < Mptr->R.
 * Output: the submatrix formed from columns Cptr[0],...,Cptr[n-1].
 */
{
	unsigned int i, j;
	MPMATm *Nptr;

	Nptr = BUILDMATm(Mptr->R, n);
	for (j = 0; j < n; j++)
		for (i = 0; i < Mptr->R; i++)
			Nptr->V[i][j] = Mptr->V[i][Cptr[j]];
	return Nptr;
}
