/* i6I.c */
/* matrix operations using MPI's */

#include <stdio.h>
#include <stdlib.h>
#include "integer.h"
#include "fun.h"

MPMATI *BUILDMATI(USI m, USI n)
/*
 * Allocates space for an m x n matrix of MPI's.
 */
{
	MPMATI *N;

	N=(MPMATI *)mmalloc(sizeof(MPMATI));
	N->R = m;
	N->C = n;
	N->V = (MPI **)mmalloc(m * n * sizeof(MPI *));
	return (N);
}

MPMATI *COPYMATI(MPMATI *Mptr)
/*
 * a replacement for the assignment *Nptr = *Mptr.
 */
{
	unsigned int i, t;
	MPMATI *Nptr;

	t = Mptr->R * Mptr->C;
	Nptr = BUILDMATI(Mptr->R, Mptr->C);
	for (i = 0; i < t; i++)
			Nptr->V[i] = COPYI(Mptr->V[i]);
	return (Nptr);
}

void FREEMATI(MPMATI *Mptr)
/*
 * frees the storage allocated to the two-dimensional array Mptr->V.
 */
{
	unsigned int i, t;

	if (Mptr == NULL)
		return;
	t = Mptr->R * Mptr->C;
	for (i = 0; i < t; i++)
		FREEMPI(Mptr->V[i]);
	ffree((char *)(Mptr->V), t * sizeof(MPI *));
	ffree((char *)Mptr, sizeof(MPMATI));
	return;
}

void SUBMI(USI p, USI q, MPI *Aptr, USI j, MPMATI *Mptr)
/*
 * subtracts *Aptr times the p-th row of *Mptr from the q-th,
 * for columns Mptr->C - 1 >= k > j.
 * 0 <= p, q <= Mptr->R - 1.
 * elt(Mptr, p, k) = 0 if k < j, elt(Mptr, p, j) = 1.
 */
{
	MPI *Y, *Temp;
	unsigned int k;

	for (k = j + 1; k <= Mptr->C - 1; k++)
	{
		Y = MULTI(Aptr, elt(Mptr, p, k));
		Temp = elt(Mptr, q, k);
		elt(Mptr, q, k) = SUBI(Temp, Y);
		FREEMPI(Temp);
		FREEMPI(Y);
	}
	return;
}

MPI *TRACEI(MPMATI *Mptr)
/*
 * *Aptr is the trace of *Mptr.
 */
{
	unsigned int i;
	MPI *Aptr, *Temp;

	Aptr = ZEROI();
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		Temp = Aptr;
		Aptr = ADDI(Aptr, elt(Mptr, i, i));
		FREEMPI(Temp);
	}
	return (Aptr);
}

MPMATI *SCALARMATI(MPMATI *Mptr, MPI *Aptr)
/*
 * returns (*Aptr) * (*Mptr).
 */
{
	unsigned int i, j;
	MPMATI *Nptr;

	Nptr = BUILDMATI(Mptr->R, Mptr->C);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		for (j = 0; j <= Nptr->C - 1; j++)
			elt(Nptr, i, j) = MULTI(elt(Mptr, i, j), Aptr);
	}
	return (Nptr);
}

MPMATI *MINUSMATI(MPMATI *Mptr)
/*
 * returns -(*Mptr).
 */
{
	unsigned int i, j;
	MPMATI *Nptr;

	Nptr = BUILDMATI(Mptr->R, Mptr->C);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		for (j = 0; j <= Nptr->C - 1; j++)
			elt(Nptr, i, j) = MINUSI(elt(Mptr, i, j));
	}
	return (Nptr);
}

MPMATI *DIAGI(MPMATI *Mptr, MPI *Aptr)
/*
 * returns *Mptr - (*Aptr) * (Unit Matrix).
 */
{
	unsigned int i;
	MPI *Temp;
	MPMATI *Nptr;

	Nptr = COPYMATI(Mptr);
	for (i = 0; i <= Mptr->R - 1; i++)
	{
			Temp = elt(Nptr, i, i);
			elt(Nptr, i, i) = SUBI(elt(Mptr, i, i), Aptr);
			FREEMPI(Temp);
	}
	return (Nptr);
}

MPMATI *ADDMATI(MPMATI *Mptr, MPMATI *Nptr)
/*
 * returns *Mptr + *Nptr.
 */
{
	unsigned int i, j;
	MPMATI *Lptr;

	Lptr = BUILDMATI(Mptr->R, Mptr->C);
	for (i = 0; i <= Lptr->R - 1; i++)
		for (j = 0; j <= Lptr->C - 1; j++)
			elt(Lptr, i, j) = ADDI(elt(Mptr, i, j), elt(Nptr, i, j));
	return (Lptr);
}

MPMATI *SUBMATI(MPMATI *Mptr, MPMATI *Nptr)
/*
 * returns *Mptr - *Nptr.
 */
{
	unsigned int i, j;
	MPMATI *Lptr;

	Lptr = BUILDMATI(Mptr->R, Mptr->C);
	for (i = 0; i <= Lptr->R - 1; i++)
		for (j = 0; j <= Lptr->C - 1; j++)
			elt(Lptr, i, j) = SUBI(elt(Mptr, i, j), elt(Nptr, i, j));
	return (Lptr);
}

MPMATI *MULTMATI(MPMATI *Mptr, MPMATI *Nptr)
/*
 * Here Mptr->C = Nptr->R. 
 * returns (*Mptr) * (*Nptr).
 */
{
	MPI *X, *Y, *Temp;
	unsigned int i, j, k;
	MPMATI *Lptr;

	Lptr = BUILDMATI(Mptr->R, Nptr->C);
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		for (j = 0; j <= Nptr->C - 1; j++)
		{
			X = ZEROI();
			for (k = 0; k <= Mptr->C - 1; k++)
			{
				Y = MULTI(elt(Mptr, i, k), elt(Nptr, k, j));
				Temp = X;
				X = ADDI(X, Y);
				FREEMPI(Temp);
				FREEMPI(Y);
			}
			elt(Lptr, i, j) = X;
		}
	}
	return (Lptr);
}

MPMATI *INVERSEMATI(MPMATI *M, MPI *N)
/*
 * If gcd(det M, N) = 1, we return INVERSE(M) (mod N), otherwise return NULL.
 */
{
	MPI *det, *G, *H, *T;
	POLYI A;
	MPMATI *output, *temp;
	unsigned int i, j;

	A = LEVERRIER_I(M, &output, &det);
	G = GCD(det, N);
	if (!EQONEI(G))
	{
		FREEMATI(output);
		output = NULL;
	}
	else
	{
		H = INVERSEM(det, N);
		temp = output;
		output = SCALARMATI(output, H);
		FREEMPI(H);
		FREEMATI(temp);
		for (i = 0; i < M->R; i++)
			for (j = 0; j < M->R; j++)
			{
				T = elt(output, i, j);
				elt(output, i, j) = MOD(T, N);
				FREEMPI(T);
			}
	}
	FREEMPI(det);
	FREEMPI(G);
	DELETE_PI(A);
	return (output);
}

MPMATI *MATR_TO_MATI(MPMATR *N)
/*
 * Converts a matrix of integer rationals to an integer matrix.
 */
{
	unsigned int i, j;
	MPMATI *temp;

	temp = BUILDMATI(N->R, N->C);
	for (i = 0; i < N->R; i++)
		for (j = 0; j < N->C; j++)
			elt(temp, i, j) = COPYI(elt(N, i, j)->N);
	return (temp);
}

MPMATR *MATI_TO_MATR(MPMATI *N)
/*
 * Converts a matrix of integers to an matrix of integer rationals;
 */
{
	unsigned int i, j;
	MPMATR *temp;

	temp = BUILDMATR(N->R, N->C);
	for (i = 0; i < N->R; i++)
		for (j = 0; j < N->C; j++)
		{
			elt(temp, i, j) = BUILDMPR();
			elt(temp, i, j)->N = COPYI(elt(N, i, j));
			elt(temp, i, j)->D = ONEI();
		}
	return (temp);
}
